// Listing8-7.S
//
// 32-bit ARM code for the
// hexadecimal string to numeric conversion.
//
// Written to run on a Raspberry Pi.

#include    "ao32aa.inc"

// The following are required to use ARM V7&8 instructions
// on Raspberry Pi devices (3, 4, & 5):

            .arch   armv8-a

// Manifest constant for tab character:

#define tab 0x9

        .data
    
// Data to convert:
    
in0:    .asciz  "0"
in1:    .asciz  "1"
in2:    .asciz  "22"
in3:    .asciz  "333"
in4:    .asciz  "4444"
in5:    .asciz  "55555"
in6:    .asciz  "666666"
in7:    .asciz  "7777777"
in8:    .asciz  "88888888"
in9:    .asciz  "9abcdef"
ina:    .asciz  "FEDCBA90"
inb:	.asciz	" ;:,\t 12345678 "
inc:	.asciz	"aAbBcCdD;"
ind:	.asciz	"eEfF_0123"
ine:	.asciz	""
inf:	.asciz	"*1234"
ing:	.asciz	"12*34"
inh:	.asciz	"1234567890"


// Pointers to strings, so asmMain can process
// all of them in a loop:
        
strPtrs:
        .word   in0, in1, in2, in3
        .word   in4, in5, in6, in7
        .word   in8, in9, ina, inb
		.word	inc, ind
        .set    inCnt, (.-strPtrs)/4    //Number of strings

        .text
        .align  2

        #include    "print.inc"        

// getTitle returns the address of a string in
// the R0 register. The c.cpp program will print
// this string before and after calling asmMain.

        .text
        .align  2
        
// Here's the title string that c.cpp will print:

title:  wastr  "Listing8-7"
        
        proc    getTitle, public
        adr     r0, title
        bx      lr
        endp    getTitle

////////////////////////////////////////////////////
//
// bufToH32
//
// Hexadecimal string to 32-bit integer conversion.
// This code allows underscores to appear within a
// hex string (but underscores cannot be the first
// character of a hex string). This code also
// allows an underscore at the end of the string;
// if you consider this to be an error, modify
// the code at label "delimiter" to test for an
// underscore in the previous character position.
//
// On entry:
//
//  R0-     Pointer to buffer holding string.
//
// On exit-
//
//  R0-     Points at the first character beyond
//          the converted string, if conversion
//          was successful.
//
//  R1-     Contains the converted value if the
//          conversion was successful.
//
//  Carry-  Clear if conversion was successful.
//          Set if unsuccessful. If unsuccessful
//          conversion, then R0 will point at
//          the character the conversion failed
//          on and R1 will contain one of the
//          following error codes:
//
//          -1: First (non-WS) character was not
//              a valid hexadecimal digit.
//
//          -2: Overflow (value did not fit into
//              32 bits).
//
//          -3: Last character was not a valid
//              delimiter character (this could
//              be considered a "soft" error).
//
//  This function will skip any leading whitespace
// (spaces or tabs) appearing at the beginning
// of the string.
        
#define delimCode 0x11
#define underscore 0x10
#define illChar 0xff

    proc    bufToH32, thumb, public
    push    {r2-r6, lr}
    
    ldr     r3, =0x10000000 // For overflow check
    ldr     r4, =classTbl   // For char classification
    
    // Skip any leading whitespace, comma, semicolon,
    // or colon characters in the string:
    
skipDelim:
    ldrb    r2, [r0], #1
    ldrb    r5, [r4, r2]    // Get classification
	cmp		r2, #0			// Check for End of String
	beq		badFirst
    cmp     r5, #delimCode  // Is it a delimiter?
    beq     skipDelim
    
    // First non-delimiter character must
    // be a valid hexadecimal digit:
    
    mov     r1, r5
    blo     nextChar        // Hexadecimal digit
            
    // If the first digit is bad (including underscore), 
    // return -1:
    
badFirst:
    sub     r0, r0, #1          // Point back at bad char
    mov     r1, #-1
    cmp     r1, r1              // Sets carry flag
    pop     {r2-r6, pc} 

    // Process hexadecimal and decimal digits.
    // Unroll the character conversion loop eight times 
    // to improve performance.
    
    .macro  doHex1
    ldrb    r2, [r0], #1
    ldrb    r5, [r4, r2]
    cmp     r5, #underscore
    beq     nextChar
    bhi     delimiter
    orr     r1, r5, r1, lsl #4
    .endm

    .macro  doHex2
    ldrb    r2, [r0], #1
    ldrb    r5, [r4, r2]
    cmp     r5, #underscore
    beq     nextChar
    bhi     delimiter
    cmp     r1, r3
    bhs     overflow
    orr     r1, r5, r1, lsl #4
    .endm

    
nextChar:
    doHex1                  // Assume no more than 8 digits.
    doHex1                  // If 8 digits or less, overflow
    doHex1                  // will not occur. Note that we've
    doHex1					// already processed the first digit.
    doHex1
    doHex1
    doHex1
    
    // If there are more than 8 digits, we need to check for
    // overflow (overflow is not guaranteed, there could be
    // leading zeros):
    
couldOverflow:
    doHex2
    doHex2
    doHex2
    doHex2
    doHex2
    doHex2
    doHex2
    doHex2
    b       couldOverflow       // If more than 16 digits.
            
    // Transfer here when encountering a legitimate
    // delimiter character (valid end of hex string).
    
delimiter:
    sub     r0, r0, #1          // Point back at delimiter
    cmp     r5, #delimCode
    bne     illegal
    adds    r0, r0, #0          // Clears carry flag
    pop     {r2-r6, pc}

    // If the first non-hex-digit character
    // is bad, return -3:
    
illegal:
    mov     r1, #-3
    cmp     r1, r1              // Sets carry flag
    pop     {r2-r6, pc}

    
overflow:
    mov     r1, #-2
    cmp     r1, r1              // Sets carry flag
    pop     {r2-r6, pc}
    endp    bufToH32

    // Classify each character:
    //
    //  0 to 0xf-   hexadecimal character
    //  0x10-       underscore
    //  0x11-       delimiter character
    //  0xff-       illegal character
    
classTbl:
    .byte   delimCode   // 0x00
    .byte   illChar     // 0x01
    .byte   illChar     // 0x02
    .byte   illChar     // 0x03
    .byte   illChar     // 0x04
    .byte   illChar     // 0x05
    .byte   illChar     // 0x06
    .byte   illChar     // 0x07
    .byte   illChar     // 0x08
    .byte   delimCode   // 0x09
    .byte   delimCode   // 0x0a
    .byte   illChar     // 0x0b
    .byte   illChar     // 0x0c
    .byte   delimCode   // 0x0d
    .byte   illChar     // 0x0e
    .byte   illChar     // 0x0f
    .byte   illChar     // 0x10
    .byte   illChar     // 0x11
    .byte   illChar     // 0x12
    .byte   illChar     // 0x13
    .byte   illChar     // 0x14
    .byte   illChar     // 0x15
    .byte   illChar     // 0x16
    .byte   illChar     // 0x17
    .byte   illChar     // 0x18
    .byte   illChar     // 0x19
    .byte   illChar     // 0x1a
    .byte   illChar     // 0x1b
    .byte   illChar     // 0x1c
    .byte   illChar     // 0x1d
    .byte   illChar     // 0x1e
    .byte   illChar     // 0x1f
    .byte   delimCode   // ' '
    .byte   illChar     // '!'
    .byte   illChar     // '"'
    .byte   illChar     // '#'
    .byte   illChar     // '$'
    .byte   illChar     // '%'
    .byte   illChar     // '&'
    .byte   illChar     // '''
    .byte   illChar     // '('
    .byte   illChar     // ')'
    .byte   illChar     // '*'
    .byte   illChar     // '+'
    .byte   delimCode   // ','
    .byte   illChar     // '-'
    .byte   illChar     // '.'
    .byte   illChar     // '/'
    .byte   0           // '0'
    .byte   1           // '1'
    .byte   2           // '2'
    .byte   3           // '3'
    .byte   4           // 'illChar'
    .byte   5           // '5'
    .byte   6           // '6'
    .byte   7           // '7'
    .byte   8           // '8'
    .byte   9           // '9'
    .byte   delimCode   // ':'
    .byte   delimCode   // ';'
    .byte   illChar     // '<'
    .byte   illChar     // '='
    .byte   illChar     // '>'
    .byte   illChar     // '?'
    .byte   illChar     // '@'
    .byte   0xa         // 'A'
    .byte   0xb         // 'B'
    .byte   0xc         // 'C'
    .byte   0xd         // 'D'
    .byte   0xe         // 'E'
    .byte   0xf         // 'F'
    .byte   illChar     // 'G'
    .byte   illChar     // 'H'
    .byte   illChar     // 'I'
    .byte   illChar     // 'J'
    .byte   illChar     // 'K'
    .byte   illChar     // 'L'
    .byte   illChar     // 'M'
    .byte   illChar     // 'N'
    .byte   illChar     // 'O'
    .byte   illChar     // 'P'
    .byte   illChar     // 'Q'
    .byte   illChar     // 'R'
    .byte   illChar     // 'S'
    .byte   illChar     // 'T'
    .byte   illChar     // 'U'
    .byte   illChar     // 'V'
    .byte   illChar     // 'W'
    .byte   illChar     // 'X'
    .byte   illChar     // 'Y'
    .byte   illChar     // 'Z'
    .byte   illChar     // '['
    .byte   illChar     // '\'
    .byte   illChar     // ']'
    .byte   illChar     // '^'
    .byte   underscore  // '_'
    .byte   illChar     // '`'
    .byte   0xa         // 'a'
    .byte   0xb         // 'b'
    .byte   0xc         // 'c'
    .byte   0xd         // 'd'
    .byte   0xe         // 'e'
    .byte   0xf         // 'f'
    .byte   illChar     // 'g'
    .byte   illChar     // 'h'
    .byte   illChar     // 'i'
    .byte   illChar     // 'j'
    .byte   illChar     // 'k'
    .byte   illChar     // 'l'
    .byte   illChar     // 'm'
    .byte   illChar     // 'n'
    .byte   illChar     // 'o'
    .byte   illChar     // 'p'
    .byte   illChar     // 'q'
    .byte   illChar     // 'r'
    .byte   illChar     // 's'
    .byte   illChar     // 't'
    .byte   illChar     // 'u'
    .byte   illChar     // 'v'
    .byte   illChar     // 'w'
    .byte   illChar     // 'x'
    .byte   illChar     // 'y'
    .byte   illChar     // 'z'
    .byte   illChar     // '{'
    .byte   illChar     // '|'
    .byte   illChar     // '}'
    .byte   illChar     // '~'
    .byte   illChar     // ''
    .byte   illChar     // 0x80
    .byte   illChar     // 0x81
    .byte   illChar     // 0x82
    .byte   illChar     // 0x83
    .byte   illChar     // 0x84
    .byte   illChar     // 0x85
    .byte   illChar     // 0x86
    .byte   illChar     // 0x87
    .byte   illChar     // 0x88
    .byte   illChar     // 0x89
    .byte   illChar     // 0x8a
    .byte   illChar     // 0x8b
    .byte   illChar     // 0x8c
    .byte   illChar     // 0x8d
    .byte   illChar     // 0x8e
    .byte   illChar     // 0x8f
    .byte   illChar     // 0x90
    .byte   illChar     // 0x91
    .byte   illChar     // 0x92
    .byte   illChar     // 0x93
    .byte   illChar     // 0x94
    .byte   illChar     // 0x95
    .byte   illChar     // 0x96
    .byte   illChar     // 0x97
    .byte   illChar     // 0x98
    .byte   illChar     // 0x99
    .byte   illChar     // 0x9a
    .byte   illChar     // 0x9b
    .byte   illChar     // 0x9c
    .byte   illChar     // 0x9d
    .byte   illChar     // 0x9e
    .byte   illChar     // 0x9f
    .byte   illChar     // 0xa0
    .byte   illChar     // 0xa1
    .byte   illChar     // 0xa2
    .byte   illChar     // 0xa3
    .byte   illChar     // 0xa4
    .byte   illChar     // 0xa5
    .byte   illChar     // 0xa6
    .byte   illChar     // 0xa7
    .byte   illChar     // 0xa8
    .byte   illChar     // 0xa9
    .byte   illChar     // 0xaa
    .byte   illChar     // 0xab
    .byte   illChar     // 0xac
    .byte   illChar     // 0xad
    .byte   illChar     // 0xae
    .byte   illChar     // 0xaf
    .byte   illChar     // 0xb0
    .byte   illChar     // 0xb1
    .byte   illChar     // 0xb2
    .byte   illChar     // 0xb3
    .byte   illChar     // 0xb4
    .byte   illChar     // 0xb5
    .byte   illChar     // 0xb6
    .byte   illChar     // 0xb7
    .byte   illChar     // 0xb8
    .byte   illChar     // 0xb9
    .byte   illChar     // 0xba
    .byte   illChar     // 0xbb
    .byte   illChar     // 0xbc
    .byte   illChar     // 0xbd
    .byte   illChar     // 0xbe
    .byte   illChar     // 0xbf
    .byte   illChar     // 0xc0
    .byte   illChar     // 0xc1
    .byte   illChar     // 0xc2
    .byte   illChar     // 0xc3
    .byte   illChar     // 0xc4
    .byte   illChar     // 0xc5
    .byte   illChar     // 0xc6
    .byte   illChar     // 0xc7
    .byte   illChar     // 0xc8
    .byte   illChar     // 0xc9
    .byte   illChar     // 0xca
    .byte   illChar     // 0xcb
    .byte   illChar     // 0xcc
    .byte   illChar     // 0xcd
    .byte   illChar     // 0xce
    .byte   illChar     // 0xcf
    .byte   illChar     // 0xd0
    .byte   illChar     // 0xd1
    .byte   illChar     // 0xd2
    .byte   illChar     // 0xd3
    .byte   illChar     // 0xd4
    .byte   illChar     // 0xd5
    .byte   illChar     // 0xd6
    .byte   illChar     // 0xd7
    .byte   illChar     // 0xd8
    .byte   illChar     // 0xd9
    .byte   illChar     // 0xda
    .byte   illChar     // 0xdb
    .byte   illChar     // 0xdc
    .byte   illChar     // 0xdd
    .byte   illChar     // 0xde
    .byte   illChar     // 0xdf
    .byte   illChar     // 0xe0
    .byte   illChar     // 0xe1
    .byte   illChar     // 0xe2
    .byte   illChar     // 0xe3
    .byte   illChar     // 0xe4
    .byte   illChar     // 0xe5
    .byte   illChar     // 0xe6
    .byte   illChar     // 0xe7
    .byte   illChar     // 0xe8
    .byte   illChar     // 0xe9
    .byte   illChar     // 0xea
    .byte   illChar     // 0xeb
    .byte   illChar     // 0xec
    .byte   illChar     // 0xed
    .byte   illChar     // 0xee
    .byte   illChar     // 0xef
    .byte   illChar     // 0xf0
    .byte   illChar     // 0xf1
    .byte   illChar     // 0xf2
    .byte   illChar     // 0xf3
    .byte   illChar     // 0xf4
    .byte   illChar     // 0xf5
    .byte   illChar     // 0xf6
    .byte   illChar     // 0xf7
    .byte   illChar     // 0xf8
    .byte   illChar     // 0xf9
    .byte   illChar     // 0xfa
    .byte   illChar     // 0xfb
    .byte   illChar     // 0xfc
    .byte   illChar     // 0xfd
    .byte   illChar     // 0xfe
    .byte   illChar     // 0xff

/////////////////////////////////////////////////////////////////////
//
// strToH32
//
// Converts the string pointed at by R0 to a numeric value
// and returns that numeric value in R1. This function
// preserves R0 (leaves it pointing at the beginning of the
// string buffer).

    proc    strToH32, thumb, public
    push    {r0, lr}
    bl      bufToH32
    pop     {r0, pc}
    endp    strToH32
            
/////////////////////////////////////////////////////////////////////
//              
// Here's the assembly language main program:
        
            proc    asmMain, public
            push    {r0-r8, lr}

            ldr     r7, =strPtrs
            mov     r6, #inCnt
convLp:     ldr     r0, [r7], #4
               
            // Convert string pointed at by R0 to 
            // a numeric value in R1 and output
            // that value:
            
            bl      strToH32
            bcc     goodConv
            
            bl      print
            wastr   "Conversion error processing \"%s\", code = %ld\n"
            b       nextConv

goodConv:
            bl      print
            wastr   "Converted '%s' to %08x\n"
                  
nextConv:
            subs    r6, #1
            bne     convLp
			
// Check for expected errors:
//
// Check empty string on input:

			ldr		r0, =ine
			bl		strToH32
			bcs		goodErr1
			bl		print
			wastr	"Expected error for empty string input\n"
			b		tryInf
			
goodErr1:	bl		print
			wastr	"Processing \"%s\" produced error code %d\n"

// Check "*1234" on input:
			
tryInf:		
			ldr		r0, =inf
			bl		strToH32
			bcs		goodErr2
			bl		print
			wastr	"Expected illegal 1st char for \"%s\"\n"
			b		tryIng
			
goodErr2:	bl		print
			wastr	"Processing \"%s\" produced error code %d\n"

// Check "12*34" on input:
			
tryIng:		
			ldr		r0, =ing
			bl		strToH32
			bcs		goodErr3
			bl		print
			wastr	"Expected illegal 1st char for \"%s\"n"
			b		tryInh
			
goodErr3:	bl		print
			wastr	"Processing \"%s\" produced error code %d\n"

// Check for overflow:
			
tryInh:		
			ldr		r0, =inh
			bl		strToH32
			bcs		goodErr4
			bl		print
			wastr	"Expected illegal 1st char for \"%s\"n"
			b		allDone
			
goodErr4:	bl		print
			wastr	"Processing \"%s\" produced error code %d\n"
			            
allDone:             
            pop     {r0-r8, pc} 
            endp    asmMain


